@@ -1,7 +1,5 @@
 /* posix_ipc.cc: POSIX IPC API for Cygwin.
 
-   Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2014, 2015 Red Hat, Inc.
-
 This file is part of Cygwin.
 
 This software is a copyrighted work licensed under the terms of the
@@ -183,9 +181,7 @@ ipc_cond_timedwait (HANDLE evt, HANDLE mtx, const struct timespec *abstime)
     ++cnt;
   if (abstime)
     {
-      if (abstime->tv_sec < 0
-	       || abstime->tv_nsec < 0
-	       || abstime->tv_nsec > 999999999)
+      if (!valid_timespec (*abstime))
 	return EINVAL;
 
       /* If a timeout is set, we create a waitable timer to wait for.
